description = """
This file contains the analytical solution to the heat equation on a square
domain.

The solutions it returns are of T(x,y,t).

dT/dt = div k grad T + f

This is a good one to converge to numerically.
"""

default_parameters = {
    'K_d': 10.0e9,
    'K_s': 100.0*10.0e9,
    'K_f':1.0e11,
    'phi':0.22,
    'k':1.0e-13,
    'eta':1.0e-3,

    'source':-1.0e5,
}

class SquareHomogenousHeatEquation():
    name = 'StaticSquarePoisson'
    space_dim = 2
    time_dep = False
    ptdim = 3
    def __init__(self,in_params=None):
        params = default_parameters.copy()
        if in_params:
            params.update(in_params)
        self.params = params
        # Yoink out the parameters
        K_d = params['K_d']
        K_s = params['K_s']
        K_f = params['K_f']
        phi = params['phi']
        k_eta = params['k']/params['eta']
        domH = 1.0
        self.source = params['source']
        # Only onle coefficent
        alpha = biot = 1.0 - K_d / K_s
        M = K_s/(alpha-phi*(1.0-K_s/K_f))
        self.k = k_eta / self.source
    def __call__(self, xt):
        pass
        return {'P':self.P(xt[:,0],xt[:,1], xt[:,2])}

class SquareForcedHeatEquation():
    name = 'StaticSquarePoisson'
    space_dim = 2
    time_dep = False
    ptdim = 3
    def __init__(self,in_params=None):
        params = default_parameters.copy()
        if in_params:
            params.update(in_params)
        self.params = params
        # Yoink out the parameters
        K_d = params['K_d']
        K_s = params['K_s']
        K_f = params['K_f']
        phi = params['phi']
        k_eta = params['k']/params['eta']
        domH = 1.0
        self.source = params['source']
        # Only onle coefficent
        alpha = biot = 1.0 - K_d / K_s
        M = K_s/(alpha-phi*(1.0-K_s/K_f))
        self.k = k_eta / self.source
    def __call__(self, xt):
        pass
        return {'P':self.P(xt[:,0],xt[:,1], xt[:,2])}

tests = [SquareHomogenousHeatEquation, SquareForcedHeatEquation]



from __future__ import print_function
import numpy as np
import scipy.stats
from collections import defaultdict

from .TestRunner import TestRunner

def rate(hs,es):
    " Calculate convergence rate "
    return scipy.stats.linregress([np.log(h) for h in hs],
                                  [np.log(e) for e in es])[0]

class ConvergenceTest(TestRunner):
    """
    Give this clas one problem and one script, and it will run a _convergence_
    test on the code, automatically.
    """
    def __init__(self, problem, script, expected_order,
                 h_path=None,
                 use_db=False,
                 **kwargs):
        self.expected_order = expected_order
        if h_path is None:
            h_path = np.linspace(0.1,2.0, 20)
        self.h_path = h_path
        self.use_db = use_db
        TestRunner.__init__(self,problem,script,**kwargs)

    def run_cases(self, h_dt_path):
        self.oracle = self.problem(self.params)
        params = self.oracle.params
        def runit(h):
            return self.script(params,h)
        if self.use_db:
            from SimDataDB import SimDataDB
            sdb = SimDataDB(self.cwd+"/conv_"+self.name+"_errors.db")
            runit = sdb.Decorate('test',[('h','FLOAT'),],
                                 [('points','ARRAY')]+[ (k,'ARRAY') for k in self.oracle.outputs ],
                                 memoize=True,dictreturn=True)(runit)

        # TODO this should be asynchronous
        self.raw = []
        for h in self.h_path:
            estimate = runit(h)
            errors = self.calc_errors(self.oracle,estimate)
            self.raw.append((h,estimate,errors))

    def analyze_cases(self):
        self.field_errors = defaultdict(lambda : [])
        for h,_,errors in self.raw:
            for k in errors:
                v = errors[k]
                self.field_errors[k].append( (h,v) )
        for k in self.field_errors.keys():
            self.field_errors[k] = np.array(self.field_errors[k])
        passed=True
        self.field_orders = {}
        for k in self.field_errors:
            es = self.field_errors[k]
            order = rate(es[:,0],es[:,1])
            self.field_orders[k] = order
            # TODO check if nan: that means exact
            if np.abs(order-self.expected_order) > 5.0e-2:
                passed=False
        return passed

    def plot_results(self):
        """ Make plots to generate a report """
        import matplotlib
        matplotlib.use('agg')
        from matplotlib import pylab as plt
        # sdb = SimDataDB(self.cwd+"/conv_"+self.oracle.name+"_errors.db")
        # Make the contour plots

        for f in self.problem.outputs:
            if self.raw[0][1][f].shape[0] == self.raw[0][1]['points'].shape[0]:
                plt.close('all')
                plt.figure()
                # Plot each sample
                for h,estimate,errors in self.raw:
                    y = estimate[f]
                    if len(y.shape)>1:
                        y = y[:,0]
                    x = estimate['points'][:,self.report_cfg['idx']]
                    ind = np.argsort(x)
                    plt.plot(x[ind],y[ind],'-+',label='h='+str(h),markevery=0.1)
                # Plot the oracle
                yo = self.oracle( estimate['points'] )[f]
                plt.plot(x,yo,'--',label='oracle')
                plt.legend()
                # Label the axes
                try:
                    plt.xlabel(self.report_cfg['x_label'])
                except KeyError:
                    plt.xlabel('x')
                try:
                    plt.ylabel(self.report_cfg['labels'][f])
                except KeyError:
                    plt.ylabel(f)
                plt.tight_layout()
                plt.savefig(self.cwd+"/"+self.name+"_"+f+"_contours.pdf")
        # Make a log log error plots
        plt.figure()
        for f in self.problem.outputs:
            plt.loglog(self.field_errors[f][:,0],
                       self.field_errors[f][:,1],'+-',label=f)
        plt.legend()
        plt.xlabel('log(h)')
        plt.ylabel('log(error)')
        plt.tight_layout()
        plt.savefig(self.cwd+"/"+self.name+"_error.pdf")


    def print_report(self):
        #for h,_,errors in self.raw:
        #    print(h,": ",errors)
        for k in self.field_orders:
            print(k,": ",self.field_orders[k])


    def test(self):
        passed = False
        self.run_cases(self.h_path)
        passed = self.analyze_cases()
        if self.report:
            self.plot_results()
            self.print_report()
        return passed



"""
Similar to the constant strain problems for mechanics, but these are for a
*sealed* poroelastic material. The fluid pressures changes in response to
the instantaneous loading. These are the static solutions, after the flow
and stresses have relaxed.

It is imperative that a poroelasticity solver pass these as ExactTests.

"""
import numpy as np

default_parameters = {
    'K_d': 10.0e9,
    'G': 10.0e9,
    'K_s': 100.0*10.0e9,
    'K_f':1.0e11,
    'phi':0.22,

    'Load':-1.0e5,
    'P_background':0.0,
}

class UndrainedUniaxial():
    name = "UndrainedUniaxial"
    space_dim = 3
    time_dep = False
    ptdim = 3
    outputs = ['sigma','U','P']
    def __init__(self, in_params=None):
        params = default_parameters.copy()
        if in_params:
            params.update(in_params)
        self.params = params
        K_d  = params['K_d']
        G    = params['G']
        K_f = params['K_f']
        K_s = params['K_s']
        phi = params['phi']
        Load = params['Load']
        P_background = params['P_background']
        alpha = 1.0-K_d/K_s
        M = K_s/(alpha-phi*(1.0-K_s/K_f))
        K_u = alpha**2*M + K_d
        H = (K_d+4.0*G/3.0)

        self.P0 = -alpha/H / ( 1.0/M + alpha**2/H) * Load + P_background
        self.eps = Load / (K_u + 4.0*G/3.0)
        sigyy_back = - P_background
        sigxx_back = - alpha*P_background - (K_d-2.0/3.0*G)/(H) * (1.0-alpha)*P_background
        self.sigyy = Load + sigyy_back
        self.sigxx = self.sigzz = (Load) * (3.0*K_u-2.0*G)/(3.0*K_u+4.0*G) + sigxx_back
        self.sigxy = 0.0
    def __call__(self, x):
        return {#'eps':np.array([[0.0,0.0,0.0],
                #                [0.0,self.eps,0.0],
                #                [0.0,0.0,0.0]]),
                'sigma':np.array([[self.sigxx,self.sigxy,0.0],
                                  [self.sigxy,self.sigyy,0.0],
                                  [0.0,  0.0,  self.sigzz]]),
                'U':np.array([ [0,self.eps*_,0] for _ in x[:,1] ]),
                'P':np.array([ self.P0 for _ in x[:,0]]),
                }

class UndrainedShear():
    name = "UndrainedShear"
    space_dim = 3
    time_dep = False
    ptdim = 3
    outputs = ['sigma','U','P']
    def __init__(self, in_params=None):
        params = default_parameters
        if in_params:
            params.update(in_params)
        self.params = params

        K_d  = params['K_d']
        G    = params['G']
        Load = params['Load']
        self.eps = Load / G
        self.sigxx = self.sigyy = self.sigzz = 0.0
        self.sigxy = Load

    def __call__(self, x):
        return {#'eps':np.array([[0.0,self.eps,0.0],
                #                [self.eps,0.0,0.0],
                #                [0.0,0.0,0.0]]),
                'sigma':np.array([[self.sigxx,self.sigxy,0.0],
                                  [self.sigxy,self.sigyy,0.0],
                                  [0.0,  0.0,  self.sigzz]]),
                'U':np.array([ [0,self.eps*_,0] for _ in x[:,0] ]),
                'P':np.zeros([x.shape[0]]),
        }

def isotropic(parameters):
    pass

def random_patch(parameters):
    pass

tests = [UndrainedUniaxial, UndrainedShear]



from . import mechanics_constant, poroelastic_constant, \
              terzaghi, poroelastic_well, single_phase_well, \
              odes, heat_equation_1d, wave_equation_1d, single_phase_well


from .mechanics_constant import Uniaxial, Shear
from .poroelastic_constant import UndrainedUniaxial, UndrainedShear
from .terzaghi import Terzaghi
from .deLeeuw import DeLeeuw
from .poroelastic_well import PoroelasticWell
from .single_phase_well import SinglePhaseWell
from .odes import Decay, Oscillator
from .heat_equation_1d import HeatEquation1D
from .wave_equation_1d import WaveEquation1D


from __future__ import print_function
import numpy as np

"""
This file contains the analytical solution to de Leeuw's consolidation
problem. It's similar to Mandel's problem, but is cylindrically symmetric.

It's a dynamic problem with one spatial dimension, radius.
The solutions it returns are P(r,t) and U(r,t).

This is a good one to converge to numerically.

Verruijt, Theory and Problems in Poroelasticity, 2013
"""

from sympy import erf, besselj, var, nsolve, pi, cos, exp, Sum, lambdify
import scipy.special as scisp

default_parameters = {
    'K_d': 10.0e9,
    'G': 10.0e9,
    'K_s': 100.0*10.0e9,
    'K_f':1.0e11,
    'phi':0.22,
    'k':1.0e-13,
    'eta':1.0e-3,

    'R':10.0,
    'Load':-1.0e5,
    'P_background':0.0,
}


class DeLeeuw():
    name = "DeLeeuw"
    space_dim = 1
    time_dep = True
    ptdim = 2
    outputs = ['P',]
    def __init__(self, in_params=None):
        params = default_parameters.copy()
        if in_params:
            params.update(in_params)
        self.params = params
        # Yoink out the parameters
        K_d = params['K_d']
        G = params['G']
        K_s = params['K_s']
        K_f = params['K_f']
        phi = params['phi']
        k = params['k']
        eta = params['eta']

        R = params['R']
        Load = params['Load']

        alpha = 1.0-K_d/K_s
        S  = phi/K_f+(alpha-phi)/K_s
        p0 = -alpha / ( alpha**2 + S*(K_d+1.0/3.0*G) ) * Load
        m  = 0.5* ( K_d+4.0*G/3.0 )/( 2.0*G )  * (alpha**2+S*(K_d+G/3.0))/(alpha**2)
        # c  = ( k/eta )/(  S+alpha**2/(K_d+4.0/3.0*G)  )
        c = (k/eta) * ( K_d+4.0*G/3.0 )/( alpha**2 + S*(K_d+4.0*G/3.0) )
        u0 = 0.5*R*Load / (alpha**2*1.0/S+K_d+G/3.0)
        uinf = 0.5*R*Load / (K_d+G/3.0)
        print('u(R,0) is ',u0,'; u(R,inf) is ',uinf, '; p(t=0) is ',p0)
        var('xi')
        J0 = lambda x : besselj(0,x)
        J1 = lambda x : besselj(1,x)
        g=(2*m*xi*J0(xi)-J1(xi))
        # xi_j = [ nsolve(g,xi,j) for j in np.linspace(0,39,14) ]
        xi_j = [float(nsolve(g,xi,0))]
        for j in range(600):
            last = xi_j[-1]
            new = nsolve(g,xi,(last+1.0e-4,last+(1.2*pi.evalf())),solver='bisect' )
            xi_j.append(float(new))
        print(xi_j)
        J0_ = lambda x : float(besselj(0,x).evalf()) # no more symbolic math
        term = lambda r,t,j : ( J0_(xi_j[j])-J0_(xi_j[j]*r/R) )/\
            ( (1-m*xi_j[j]**2-1.0/(4*m) )*J0_(xi_j[j]) ) \
            * exp(-xi_j[j]**2*c*t/R**2)
        def P_sym(r,t):
            return np.array([
                sum([ float(p0*term(r_,t_,i)) for i in range(len(xi_j))])
                for r_,t_ in zip(r,t) ]) + params['P_background']
        term_scipy = lambda r,t,j : ( scisp.j0(xi_j[j])-scisp.j0(xi_j[j]*r/R) )/\
            ( (1-m*xi_j[j]**2-1.0/(4*m) )*scisp.j0(xi_j[j]) ) \
            * np.exp(-xi_j[j]**2*c*t/R**2)
        def P_sci(r,t):
            return sum([ (p0*term_scipy(r,t,i)) for i in range(len(xi_j))]) + params['P_background']
        def U(r,t):
            # TODO put in the u solution. Is there one?
            return u0*r/R*np.ones(r.shape)
        # from IPython import embed ; embed()
        self.P = P_sci
        self.U = U
    def __call__(self, xt):
        return {'P':self.P(xt[:,0],xt[:,1])}
                # 'U':self.U(xt[:,0],xt[:,1]) }

tests = [DeLeeuw]



# Parent class
#import unittest as ut
import numpy as np
import os

class TestRunner(): #ut.TestCase):
    """
    Give this clas one problem and one script, and it will run a _convergence_
    test on the code, automatically.
    """
    def __init__(self, problem, script, params=None,
                 scratch_space='./detest_report', extra_name='', report=False,report_cfg={}):
        self.problem = problem
        self.script = script
        self.name = "exact_"+problem.name+('_'+extra_name+'_' if extra_name else '')
        self.params = params
        self.report = False
        if scratch_space:
            self.cwd = scratch_space
        else:
            self.cwd = '.'
        os.makedirs(self.cwd, exist_ok=True)
        self.report_cfg = {
            'idx':-1,
        }
        self.report_cfg.update(report_cfg)
        
    def calc_errors(self, oracle, estimate):
        errors = {}
        orc = oracle(estimate['points'])
        # regularized = lambda x : x if x > 1.0e-8 else 1.0e-8
        regularized = lambda x : x if x > 0.0 else 1.0
        for field in orc.keys():
            expected = orc[field]
            e = np.linalg.norm(estimate[field].ravel() - expected.ravel())\
                / regularized(np.linalg.norm(expected))
            # e = np.linalg.norm(estimate[field].ravel()-expected.ravel()) / float(len(expected))
            errors[field] = e
        # from IPython import embed ; embed()
        return errors

    def test(self):
        return False



"""
Constant flow production from an infinitely large reservoir, with poromechanics.


"""

import numpy as np
from scipy.special import exp1

default_parameters = {
    'K_d': 10.0e9,
    'G': 10.0e9,
    'K_s': 100.0*10.0e9,
    'K_f':1.0e11,
    'phi':0.22,
    'k':1.0e-13,
    'eta':1.0e-3,

    'Q':-1.0e5,
    'P_background' : 9.5e6,
}

class PoroelasticWell():
    name = "PoroelasticWell"
    space_dim = 1
    time_dep = True
    ptdim = 2
    outputs = ['P','U']
    def __init__(self,in_params=None):
        params = default_parameters.copy()
        if in_params:
            params.update(in_params)
        self.params = params
        # Yoink out the parameters
        K_d = params['K_d']
        G = params['G']
        K_s = params['K_s']
        K_f = params['K_f']
        phi = params['phi']
        k_eta = params['k']/params['eta']

        P_background = params['P_background']
        Q_s = params['Q']
        alpha = 1.0 - K_d/K_s

        M = K_s/(alpha-phi*(1.0-K_s/K_f))
        # K_u = alpha**2*M + K_d
        # K_u = (K_d + alpha**2*M)  * 12.0*K_d/(21.0*K_d-9.0*alpha**2*M) # Mark's
        K_u = K_d * (12*alpha**2*M + 21*K_d)/( 21*K_d - 9*alpha**2*M)
        H = K_u + 4.0*G/3.0
        D = k_eta*(M - alpha**2 * M**2 /H)
        # D = k_eta * M * (K_d+4.0*G/3.0) / (K_u+4.0*G/3)
        def U(r,t):
            xi = r**2/(4.0*D*t)
            f_xi = (1 - np.exp(-xi) ) / xi + exp1(xi)
            return Q_s * alpha * f_xi * r / ( 8.0*np.pi*k_eta* (K_d + 4.0*G/3.0) )

        def P(r,t):
            xi = r**2/(4.0*D*t)
            return Q_s / (4.0*np.pi*k_eta) * exp1(xi) + P_background

        self.U = U
        self.P = P

    def __call__(self, xt):
        return {'P':self.P(xt[:,0],xt[:,1]),
                'U':self.U(xt[:,0],xt[:,1])}

tests = [ PoroelasticWell ]



"""
It works by populating the unittest framework.

This will churn through all of the numerical tests. It is very expensive.

It can operate in a randomized testing fashion to minimize the cost
of testing incremental updates. That is, you can configure it to do a
subsample set frequently, and then schedule the exhaustive tests less
frequently.
"""

from .ExactTest import ExactTest
from .ConvergenceTest import ConvergenceTest

import unittest as ut

def make_suite(suite, cwd='./detest_report/', report=False):
    # Inject the current working directory into them
    for e in suite:
        e.cwd = cwd
        e.report = report
    # A make a unittest classmethod
    def make_test(test):
        def fn(self):
            return self.assertTrue(test())
        return fn
    # Fill out a new type and return it
    attrs = {
        'test_'+e.name : make_test(e.test) for e in suite
    }
    return type('MySuite', (ut.TestCase,), attrs)



import numpy as np

description = """
This file contains the analytical solution to the square Poisson problem.

The solutions it returns are P(x,y).

This is a good one to converge to numerically.
"""

default_parameters = {
    'K_d': 10.0e9,
    'K_s': 100.0*10.0e9,
    'K_f':1.0e11,
    'phi':0.22,
    'k':1.0e-13,
    'eta':1.0e-3,

    'source':-1.0e5,
}

class SquarePoisson():
    name = 'StaticSquarePoisson'
    space_dim = 2
    time_dep = False
    ptdim = 3
    def __init__(self,params=default_parameters):
        self.params = params.copy()
        # Yoink out the parameters
        K_d = params['K_d']
        K_s = params['K_s']
        K_f = params['K_f']
        phi = params['phi']
        k_eta = params['k']/params['eta']
        domH = 1.0
        self.source = params['source']
        # Only onle coefficent
        alpha = biot = 1.0 - K_d / K_s
        M = K_s/(alpha-phi*(1.0-K_s/K_f))
        self.k = k_eta / self.source
    def __call__(self, xt):
        k = self.k
        panal = np.empty(X.shape[0])
        t = lambda k,x,y : np.sin(k*np.pi*(1.0+x)/2.0)/( k*k*k * np.sinh( k*np.pi )) * \
            ( np.sinh( k*np.pi*(1.0+y)/2.0 ) + np.sinh( k*np.pi*(1.0-y)/2.0) )
        for i,y in enumerate(X):
            panal[i] = (1.0-xt[0]**2.0)/2.0 - 16.0/(np.pi)**3 * \
                sum([t(k,yt[0],yt[1]) for k in xrange(1,101,2) ])
        return {'P':panal}


tests = [SquarePoisson]



import numpy as np

name = "terzaghi"

description = """
This file contains the analytical solution to Terzaghi's consolidation problem.

It's a dynamic problem with one spatial dimension, height.
The solutions it returns are P(z,t) and U(z,t).

This is a good one to converge to numerically.
"""

default_parameters = {
    'K_d': 10.0e9,
    'G': 10.0e9,
    'K_s': 100.0*10.0e9,
    'K_f':1.0e11,
    'phi':0.22,
    'k':1.0e-13,
    'eta':1.0e-3,

    'H':100.0,
    'Load':-1.0e5,
    'P_background':0.0,
}

class Terzaghi():
    name = "Terzaghi"
    space_dim = 1
    time_dep = True
    ptdim = 2
    outputs = ['P','U']
    def __init__(self,in_params=None):
        params = default_parameters.copy()
        if in_params:
            params.update(in_params)
        self.params = params
        # Yoink out the parameters
        K_d = params['K_d']
        G = params['G']
        K_s = params['K_s']
        K_f = params['K_f']
        phi = params['phi']
        k_eta = params['k']/params['eta']
        P_back = params['P_background']
        domH = params['H']
        Load = params['Load']

        # Poisson Ratio
        nu = (3.0*K_d - 2.0*G)/(6.0*K_d+2.0*G)

        alpha = biot = 1.0 - K_d / K_s
        Q = K_d/alpha
        H = K_d+4.0*G/3.0
        M = K_s/(alpha-phi*(1.0-K_s/K_f))
        K_u = alpha**2*M + K_d
        a = 1.0/H#( 1.0 - 2.0*nu) / ( 2.0*G*(1.0-nu) )
        ai = 1.0/(K_u + 4.0*G/3.0)

        p0 = -alpha/H / ( 1.0/M + alpha**2/H) * Load
        Cf = (k_eta) / ( a * alpha**2.0 + 1.0/M )
        uinf = a * Load * domH
        pressure_mass = biot/K_d

        def P(z,t):
            term = lambda m, z,t : 1.0/(2.0*m+1.0) \
                   * np.exp( - ((2.0*m+1.0)*np.pi/(2.0*domH))**2 * Cf * t ) \
                   * np.sin( (2.0*m+1.0)*np.pi/(2.0*domH) * z )
            return P_back + 4.0/np.pi* p0 * np.sum( [term(m, z,t) for m in np.arange(0.0,30.0) ],axis=0)

        def U(z,t):
            term = lambda m, z,t : 1.0/(2.0*m+1.0)**2 \
                   * np.exp( - ((2.0*m+1.0)*np.pi/(2.0*domH))**2 * Cf * t ) \
                   * np.cos( (2.0*m+1.0)*np.pi/(2.0*domH) * z )
            return a*Load*(domH-z) - 8.0/(np.pi**2) * (a-ai)*(domH)*Load\
                * np.sum( [term(m, z,t) for m in np.arange(0.0,30.0) ], axis=0)
        self.P = P
        self.U = U
    def __call__(self, xt):
        return {'P':self.P(xt[:,0],xt[:,1]),
                'U':self.U(xt[:,0],xt[:,1])}


tests = [Terzaghi]



"""
Westergard's solution for a thin crack in tension.

Returns a displacment that's functions of x,y, with the crack at the center.

The trick to this problem is to use the analytical solution as the outer
dirrichlet boundary conditions because the true BCs are at infinity.

Note: this problem is discontinuous across y=0, x in [-c,c].
"""

parameters = {
    'K': 10.0e9,
    'G': 10.0e9,

    'W':1.0,
    'Lcrack': 0.1,
    'Load':10.0,
}

class ThinCrack():
    name = "ThinCrack"
    space_dim = 2
    time_dep = False
    ptdim = 2
    def __init__(self,params):
        # TODO Migrate from periflakes
        def U(x,u):
            pass
        self.U = U
    def __call__(self,x):
        return {'U':U}



"""
These are constant strain problems for mechanics.

A finite element code will solve these to within machine precision. Codes
based on non-standard "theories" or models will not necessarily capture
these. It is imperative that a mechanics solver pass these as ExactTests.
"""

import numpy as np

default_parameters = {
    'K': 10.0e9,
    'G': 10.0e9,

    'Load':-1.0e5,
}

class Uniaxial():
    name = "mechanics_constant_Uniaxial"
    space_dim = 3
    time_dep = False
    ptdim = 3
    outputs = ['sigma','U']
    def __init__(self, in_params=None):
        params = default_parameters.copy()
        if in_params:
            params.update(in_params)
        self.params = params
        K_d  = params['K']
        G    = params['G']
        Load = params['Load']
        self.eps = Load / (K_d+4.0*G/3.0)
        self.sigyy = Load
        self.sigxx = self.sigzz = Load * (3.0*K_d-2.0*G)/(3.0*K_d+4.0*G)
        self.sigxy = 0.0
    def __call__(self, x):
        return {#'eps':np.array([[0.0,0.0,0.0],
                #                [0.0,self.eps,0.0],
                #                [0.0,0.0,0.0]]),
                'sigma':np.array([[self.sigxx,self.sigxy,0.0],
                                  [self.sigxy,self.sigyy,0.0],
                                  [0.0,  0.0,  self.sigzz]]),
                'U':np.array([ [0,self.eps*_,0] for _ in x[:,1] ])
                }

class Shear():
    name = "mechanics_constant_Shear"
    space_dim = 3
    time_dep = False
    ptdim = 3
    outputs = ['sigma','U']
    def __init__(self, in_params=None):
        params = default_parameters.copy()
        if in_params:
            params.update(in_params)
        self.params = params
        K_d  = params['K']
        G    = params['G']
        Load = params['Load']
        self.eps = Load / G
        self.sigxx = self.sigyy = self.sigzz = 0.0
        self.sigxy = Load

    def __call__(self, x):
        return {#'eps':np.array([[0.0,self.eps,0.0],
                #                [self.eps,0.0,0.0],
                #                [0.0,0.0,0.0]]),
                'sigma':np.array([[self.sigxx,self.sigxy,0.0],
                                  [self.sigxy,self.sigyy,0.0],
                                  [0.0,  0.0,  self.sigzz]]),
                'U':np.array([ [0,self.eps*_,0] for _ in x[:,0] ])
        }



class isotropic():
    def __init__(self,params):
        pass
    def __call__(self,x,y,z):
        return {}
class random_patch():
    def __init__(self,params):
        pass
    def __call__(self,x,y,z):
        return {}


# A list of the ones in here in case you want to loop
tests = [Uniaxial, Shear]



"""
Constant flow production from an infinitely large reservoir, without poromechanics.


"""

import numpy as np
from scipy.special import exp1

default_parameters = {
    'K_s': 100.0*10.0e9,
    'K_f':2.24e9,
    'phi':0.22,
    'k':1.0e-13,
    'eta':1.0e-3,

    'Q':-1.0e5,
    'P_background' : 9.5e6,
}

class SinglePhaseWell():
    name = "SinglePhaseWell"
    space_dim = 1
    time_dep = True
    ptdim = 2
    default_params = default_parameters
    outputs = ['P']
    def __init__(self,in_params=None):
        params = self.default_params.copy()
        if in_params:
            params.update(in_params)
        self.params = params
        # Yoink out the parameters
        K_s = params['K_s']
        K_f = params['K_f']
        phi = params['phi']
        k_eta = params['k']/params['eta']

        P_background = params['P_background']
        Q_s = params['Q']
        #alpha = 1.0 - K_d/K_s

        M = K_f/phi #?
        D = k_eta*M

        # TODO This is wrong
        def P(r,t):
            xi = r**2/(4.0*D*t)
            return Q_s / (4.0*np.pi*k_eta) * exp1(xi) + P_background

        self.P = P

    def __call__(self, xt):
        return {'P':self.P(xt[:,0],xt[:,1])}


class SteadySinglePhaseWell():
    name = "SteadySinglePhaseWell"
    space_dim = 1
    time_dep = False
    ptdim = 1
    default_params = default_parameters
    outputs = ['P']
    def __init__(self,in_params=None):
        params = self.default_params.copy()
        if in_params:
            params.update(in_params)
        # Yoink out the parameters
        K_d = params['K_d']
        G = params['G']
        K_s = params['K_s']
        K_f = params['K_f']
        phi = params['phi']
        k_eta = params['k']/params['eta']

        P_background = params['P_background']
        Q_s = params['Q']
        alpha = 1.0 - K_d/K_s

        M = K_s/(alpha-phi*(1.0-K_s/K_f))
        K_u = alpha**2*M + K_d
        H = K_u + 4.0*G/3.0
        D = k_eta*(M - alpha**2 * M**2 /H)

        # TODO This is wrong
        def P(r):
            xi = r**2/(4.0*D*t)
            return Q_s / (4.0*np.pi*k_eta) * exp1(xi) + P_background

        self.P = P

    def __call__(self, xt):
        return {'P':self.P(xt[:,0],xt[:,1])}


tests = [ SinglePhaseWell ]



from __future__ import print_function
import glob, re
import vtk
from vtk.util.numpy_support import vtk_to_numpy

def list_files(fpattern):
    files = glob.glob(fpattern)
    grab_digit = lambda f : int(re.search("([0-9]*)\.vtk",f).groups()[-1])
    files.sort(key=lambda f: grab_digit(f) )
    return files

def iterate(fpattern):
    files = list_files(fpattern)
    for f in files:
        yield get_data(f)

def get_data(fname):
    r=vtk.vtkUnstructuredGridReader()
    r.SetFileName(fname)
    r.ReadAllFieldsOn()
    r.ReadAllScalarsOn()
    r.ReadAllVectorsOn()
    r.Update()
    op = r.GetOutput()
    x = vtk_to_numpy( op.GetPoints().GetData() )
    pd = op.GetPointData()
    n = pd.GetNumberOfArrays()
    dic = { pd.GetArrayName(i):vtk_to_numpy( pd.GetArray(i) ) 
            for i in range(n) }
    cd = op.GetCellData()
    nc = cd.GetNumberOfArrays()
    dic.update( { 'cell_'+cd.GetArrayName(i):vtk_to_numpy( cd.GetArray(i) ) 
                    for i in range(nc) } )
    return x,dic

def cat_with_times(fpattern,timefile, fields):
    with open(timefile,"r") as f:
        times = [ float(l) for l in f ]
    for dat in iterate(fpattern):
        pass



from __future__ import print_function

description = """
This file contains the analytical solution to the heat equation on a line
domain from [0,1].

The solutions it returns are of u(x,t) and v(x,t).

d^2u/dt^2 =  k d^2/dx^2 u + f
u(0,t) = 0
u(1,t) = 0
u(x,0) = u0(x)

The parameters f and u0 are functions of x; this class does the work in sympy 
to generate the solution from the Green's funciton.
This is a good one to converge to numerically.
"""

import sympy
import numpy as np
default_parameters = {
    'k':1.0,
    'f': lambda x : 0.0,
    'u0': lambda x : 4*x*(1-x),
    'v0': lambda x : 0,
}

class WaveEquation1D():
    __doc__ = description
    name = 'WaveEquation1D'
    space_dim = 1
    time_dep = True
    ptdim = 2
    outputs = ['u','v']
    def __init__(self,in_params=None):
        params = default_parameters.copy()
        if in_params:
            params.update(in_params)
        self.params = params
        # Yoink out the parameters
        self.k = params['k']
        u0 = params['u0']
        f = params['f']
        n,xi = sympy.symbols('n xi',positive=True)
        green1 = sympy.integrate(sympy.sin(n*sympy.pi*xi)*u0(xi),(xi,0,1))
        term1 = sympy.lambdify([n],green1)
        self.terms = [ term1(n) for n in range(1,21) ]
        print(self.terms)
    def __call__(self, xt):
        tot_u = np.zeros(xt.shape[0])
        tot_v = np.zeros(xt.shape[0])
        for nminus1,tn in enumerate(self.terms):
            n = nminus1+1
            tot_u += 2.0*np.sin(n*np.pi*xt[:,0])*np.cos(self.k*n*np.pi*xt[:,1])*tn
            tot_v += -self.k*np.pi* 2.0*np.sin(n*np.pi*xt[:,0])*np.sin(self.k*n*np.pi*xt[:,1])*tn
        return {'u':tot_u,'v':tot_v}

tests = [WaveEquation1D]



from __future__ import print_function
from .TestRunner import TestRunner

class ExactTest(TestRunner):
    """
    Give this clas one problem and one script, and it will run it with a
    bunch of permutations and assert that the error is within machine
    precision. This class uses the standard unittest framework because
    the tests should be _cheap_.

    There is an optional rigor parameter which controls just how far it will
    go.
    """
    def __init__(self, problem, script, tolerance=1.0e-12,
                params = None, extra_name='',report=False,
                scratch_space=''):
        self.tolerance = tolerance
        TestRunner.__init__(self,problem,script,
                            params=params,
                            scratch_space=scratch_space,
                            extra_name=extra_name)

    def test(self):
        passed = True
        # TODO: randomized testing of parameters
        oracle = self.problem(self.params)
        params = oracle.params
        def runit(h):
            ans = self.script(params,h)
            errors = self.calc_errors(oracle,ans)
            return ans, errors
        estimate, errors = runit(1.0)
        for k in errors:
            v = errors[k]
            if v > self.tolerance:
                print("Failed test key ",k,": error was ",v,".")
                passed = False
        return passed



import numpy as np

name = "odes"

description = """
This file contains some basic ODEs with analytical solutions
"""

default_parameters = {
'x0':1.0,
'v0':0.0,
'u0':0.0,
'm':1.0,
'k':2.0,
'f':3.0,
'T_max':10.0
}

class Decay():
    """
    du / dt = - k u
    """
    name = "Decay"
    space_dim = 0
    time_dep = True
    ptdim = 1
    outputs = ['u']
    default_parameters = {
        'u0':1.0,
        'T_max':10.0,
        'k':1.0
    }
    def __init__(self,in_params=None):
        self.params = self.default_parameters.copy()
        if in_params: self.params.update(in_params)
    def __call__(self,t):
        u = self.params['u0']*np.exp( -self.params['k']*t)
        return {'u':u}

class Oscillator():
    """
    Classical harmonic oscillator

    dx / dt = v
    m dv / dt = -k x
    """
    name = "Oscillator"
    space_dim = 0
    time_dep = True
    ptdim = 1
    outputs = ['x','v']
    default_parameters = {
        'x0':1.0,'v0':0.0, 'T_max':10.0,
        'm':1.0,'k':2.0,'f':0.0,
    }
    def __init__(self,in_params=None):
        self.params = self.default_parameters.copy()
        if in_params: self.params.update(in_params)
        self.omega = np.sqrt(self.params['k'] / self.params['m'])
        self.amplitude = self.params['x0'] # TODO
        self.phase = 0.0

    def __call__(self,t):
        x = self.amplitude * np.cos(self.omega * t + self.phase)
        v = - self.amplitude*self.omega * np.sin(self.omega * t + self.phase)
        return {'x':x,'v':v}

class ElectricCurrentDAE():
    """

    Watch out: This one is not a normal ODE
    """
    name = "ElectricCurrentDAE"
    space_dim = 0
    time_dep = True
    ptdim = 1
    outputs = ['x','v','T','I']
    def __init__(self,params=default_parameters):
        self.params = params
    def __call__(self,t):
        x = t.copy()
        v = t.copy()
        T = t.copy()
        I = t.copy()
        return {'x':x,'v':v,'T':T,'I':I}

tests = [Decay, Oscillator, ElectricCurrentDAE]



import numpy as np
from scipy.special import ellipj, ellipk

description = """
This is the traditional nonlinear pendulum with R=1.
"""
class PendulumTheta():
    """
    The fully nonlinear pendulum, in theta

    d^theta / dt^2 = g sin(theta)
    """
    name = "PendulumTheta"
    space_dim = 0
    time_dep = True
    ptdim = 1
    outputs = ['theta','v']
    default_parameters = {
        'theta0':1.0,'v0':0.0,
        'g':9.81,
    }
    def __init__(self,in_params=None):
        self.params = self.default_parameters.copy()
        if in_params: self.params.update(in_params)
        self.g = self.params['g']
        
    def __call__(self,t):
        S = np.sin(0.5*( self.theta0) )
        K_S = ellipk(S**2)
        omega_0 = np.sqrt(9.81)
        sn,cn,dn,ph = ellipj( K_S - omega_0*t, S**2 )
        theta = 2.0*np.arcsin( S*sn )
        d_sn_du = cn*dn
        d_sn_dt = -omega_0 * d_sn_du
        d_theta_dt = 2.0*S*d_sn_dt / np.sqrt(1.0-(S*sn)**2)
        return {'theta':theta,
                'v':d_theta_dt}

class PendulumXY():
     """
    The fully nonlinear pendulum, in x,y

    d^theta / dt^2 = g sin(theta)
    """
    name = "PendulumXY"
    space_dim = 0
    time_dep = True
    ptdim = 1
    outputs = ['x','y','vx','vy']
    default_parameters = {
        'theta0':1.0,'v0':0.0, 'T_max':10.0,
        'g':9.81,
    }
    def __init__(self,in_params=None):
        self.params = self.default_parameters.copy()
        if in_params: self.params.update(in_params)
        self.g = self.params['g']
        
    def __call__(self,t):
        S = np.sin(0.5*( self.theta0) )
        K_S = ellipk(S**2)
        omega_0 = np.sqrt(9.81)
        sn,cn,dn,ph = ellipj( K_S - omega_0*t, S**2 )
        theta = 2.0*np.arcsin( S*sn )
        d_sn_du = cn*dn
        d_sn_dt = -omega_0 * d_sn_du
        d_theta_dt = 2.0*S*d_sn_dt / np.sqrt(1.0-(S*sn)**2)
        return {'x':np.cos(theta),
                'y':np.sin(theta),
                'vx':-d_theta_dt*np.sin(theta),
                'vy': d_theta_dt*np.cos(theta)}

tests = [PendulumXY, PendulumTheta]



import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "detest",
    version = "0.2",
    author = "Alejandro Francisco Queiruga",
    description = "A differential equation testing suite",
    url="https://github.com/afqueiruga/detest",
    license = "GNU LGPL",
    keywords = "",
    packages = find_packages(exclude=['test']),
    test_suite='test',
    long_description=read('Readme.md'),
    long_description_content_type='text/markdown',
    classifiers=[],
)
